!!******************************************************************************
!!
!!  This file is part of the AMUN source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations on uniform or
!!  adaptive mesh.
!!
!!  Copyright (C) 2020-2023 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: COMPRESSION
!!
!!  This module provides compression for the XML-binary format.
!!
!!******************************************************************************
!
module compression

  implicit none

! interfaces to compression algorithms
!
#ifdef ZSTD
  interface

    integer(c_size_t) function zstd_bound(srcSize) &
                               bind(C, name="ZSTD_compressBound")
      use iso_c_binding, only: c_size_t
      implicit none
      integer(kind=c_size_t), value :: srcSize
    end function zstd_bound

    integer(c_size_t) function zstd_compress(dst, dstCapacity, &
                                             src, srcSize, level) &
                               bind(C, name="ZSTD_compress")
      use iso_c_binding, only: c_size_t, c_int, c_ptr
      implicit none
      integer(kind=c_size_t), value :: srcSize, dstCapacity
      type(c_ptr)           , value :: src, dst
      integer(kind=c_int)   , value :: level
    end function zstd_compress

    integer(kind=c_int) function zstd_iserror(code) bind(C, name="ZSTD_isError")
      use iso_c_binding, only: c_int, c_size_t
      implicit none
      integer(kind=c_size_t), value :: code
    end function zstd_iserror

  end interface
#endif /* ZSTD */
#ifdef LZ4
  interface

    integer(kind=c_size_t) function lz4_bound(srcSize, prefsPtr)               &
                                    bind(C, name="LZ4F_compressFrameBound")
      use iso_c_binding, only: c_size_t, c_ptr
      implicit none
      integer(kind=c_size_t), value :: srcSize
      type(c_ptr)           , value :: prefsPtr
    end function lz4_bound

    integer(kind=c_size_t) function lz4_compress(dst, dstCapacity,             &
                                    src, srcSize, prefsPtr)                    &
                                    bind(C, name="LZ4F_compressFrame")
      use iso_c_binding, only: c_size_t, c_ptr
      implicit none
      integer(kind=c_size_t), value :: dstCapacity, srcSize
      type(c_ptr)           , value :: src, dst, prefsPtr
    end function lz4_compress

    integer(kind=c_int) function lz4_iserror(code) bind(C, name="LZ4F_isError")
      use iso_c_binding, only: c_int, c_size_t
      implicit none
      integer(kind=c_size_t), value :: code
    end function lz4_iserror

  end interface
#endif /* LZ4 */
#ifdef LZMA
  interface

    integer(c_int) function lzma_compress(preset, check, allocator,            &
                                          src, srcSize,                        &
                                          dst, dstPos, dstCapacity)            &
                            bind(C, name="lzma_easy_buffer_encode")
      use iso_c_binding, only: c_int, c_size_t, c_ptr
      implicit none
      integer(c_int)        , value :: preset, check
      integer(kind=c_size_t), value :: srcSize, dstCapacity
      type(c_ptr)           , value :: allocator, src, dst, dstPos
    end function lzma_compress

  end interface
#endif /* LZMA */

! supported compressors
!
  enum, bind(c)
    enumerator compressor_none
    enumerator compressor_zstd
    enumerator compressor_lz4
    enumerator compressor_lzma
  end enum

! supported data encoders
!
  enum, bind(c)
    enumerator encoder_none
    enumerator encoder_shuffle
    enumerator encoder_bytedelta
  end enum

  private

  public :: check_compressor, is_compressor_on
  public :: get_compressor_name, get_compressed_file_suffix
  public :: compression_bound, compress
  public :: check_encoder, is_encoder_on, get_encoder_name, encode

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!!
!!***  PUBLIC SUBROUTINES  ****************************************************
!!
!===============================================================================
!
!===============================================================================
!
! subroutine CHECK_COMPRESSOR:
! ---------------------------
!
!   Subroutine sets the compressor ID and the compression level based on
!   the compressor name.
!
!   Arguments:
!
!     compressor_name   - the compressor name;
!     compressor_id     - the compressor ID;
!     compression_level - the compression level;
!
!===============================================================================
!
  subroutine check_compressor(compressor_name, compressor_id, &
                                               compression_level)

    implicit none

    character(len=*) , intent(in   ) :: compressor_name
    integer          , intent(  out) :: compressor_id
    integer          , intent(inout) :: compression_level

!-------------------------------------------------------------------------------
!
    select case(trim(adjustl(compressor_name)))
#ifdef ZSTD
    case("zstd", "ZSTD", "zst", "ZST", "zstandard", "Zstandard")
      compressor_id = compressor_zstd
      if (compression_level < -19 .or. &
          compression_level >  19) compression_level = 3
#endif /* ZSTD */
#ifdef LZ4
    case("lz4", "LZ4", "lz4hc", "LZ4HC")
      compressor_id = compressor_lz4
      if (compression_level < -12 .or. &
          compression_level >  12) compression_level = 0
#endif /* LZ4 */
#ifdef LZMA
    case("lzma", "LZMA", "xz", "XZ")
      compressor_id = compressor_lzma
      if (compression_level <   0 .or. &
          compression_level >   9) compression_level = 6
#endif /* LZMA */
    case default
      compressor_id = compressor_none
    end select

!-------------------------------------------------------------------------------
!
  end subroutine check_compressor
!
!===============================================================================
!
! function IS_COMPRESSOR_ON:
! -------------------------
!
!   The function determines if any compression is used.
!
!   Arguments:
!
!     compressor_id - the compressor ID;
!
!===============================================================================
!
  logical function is_compressor_on(compressor_id)

    implicit none

    integer, intent(in) :: compressor_id

!-------------------------------------------------------------------------------
!
    is_compressor_on = compressor_id /= compressor_none
    return

!-------------------------------------------------------------------------------
!
  end function is_compressor_on
!
!===============================================================================
!
! function GET_COMPRESSOR_NAME:
! ----------------------------
!
!   Function returns the name of the compressor.
!
!   Arguments:
!
!     compressor_id - the compressor ID;
!
!===============================================================================
!
  character(len=4) function get_compressor_name(compressor_id)

    implicit none

    integer, intent(in) :: compressor_id

!-------------------------------------------------------------------------------
!
    select case(compressor_id)
    case(compressor_zstd)
      get_compressor_name = 'zstd'
    case(compressor_lz4)
      get_compressor_name = 'lz4'
    case(compressor_lzma)
      get_compressor_name = 'lzma'
    case default
      get_compressor_name = 'none'
    end select
    return

!-------------------------------------------------------------------------------
!
  end function get_compressor_name
!
!===============================================================================
!
! function GET_COMPRESSED_FILE_SUFFIX:
! -----------------------------------
!
!   Function returns the filename suffix for the compressor.
!
!   Arguments:
!
!     compressor_id - the compressor ID;
!
!===============================================================================
!
  character(len=4) function get_compressed_file_suffix(compressor_id)

    implicit none

    integer, intent(in) :: compressor_id

!-------------------------------------------------------------------------------
!
    select case(compressor_id)
    case(compressor_zstd)
      get_compressed_file_suffix = '.zst'
    case(compressor_lz4)
      get_compressed_file_suffix = '.lz4'
    case(compressor_lzma)
      get_compressed_file_suffix = '.xz'
    case default
      get_compressed_file_suffix = ''
    end select
    return

!-------------------------------------------------------------------------------
!
  end function get_compressed_file_suffix
!
!===============================================================================
!
! function COMPRESSION_BOUND:
! --------------------------
!
!   Function returns the minimum buffer size required to perform
!   the compression.
!
!   Arguments:
!
!     compressor_id     - the compressor ID;
!     compression_level - the compression level;
!     ilen              - the length of the sequence of bytes to compress;
!
!===============================================================================
!
  integer(kind=8) function compression_bound(compressor_id, &
                                             compression_level, ilen)

    use iso_c_binding, only: c_loc

    implicit none

    integer        , intent(in) :: compressor_id, compression_level
    integer(kind=8), intent(in) :: ilen

#ifdef LZ4
    integer, dimension(14), target :: prefs = [5, 0, 1, 0, 0, 0, 0, 0, &
                                                           0, 0, 0, 0, 0, 0]
#endif /* LZ4 */

!-------------------------------------------------------------------------------
!
    select case(compressor_id)
#ifdef ZSTD
    case(compressor_zstd)
      compression_bound = zstd_bound(ilen)
#endif /* ZSTD */
#ifdef LZ4
    case(compressor_lz4)
      prefs(5:6) = transfer(ilen, [ 0_4 ])
      prefs(9) = compression_level
      compression_bound = lz4_bound(ilen, c_loc(prefs))
#endif /* LZ4 */
    case default
      compression_bound = ilen
    end select

    return

!-------------------------------------------------------------------------------
!
  end function compression_bound
!
!===============================================================================
!
! subroutine COMPRESS:
! -------------------
!
!   Subroutine compresses the input buffer.
!
!   Arguments:
!
!     compressor_id     - the compressor ID;
!     compression_level - the compression level;
!     input             - the pointer to the input sequence of bytes;
!     ilen              - the length of input;
!     buffer            - the buffer where the compression is done;
!     clen              - the length of buffer in bytes; once the compression
!                         was successful, it returns the length of
!                         the compressed stream;
!
!===============================================================================
!
  subroutine compress(compressor_id, compression_level, &
                                                    input, ilen, buffer, clen)

    use iso_c_binding, only : c_int, c_loc, c_ptr
#ifdef LZ4
    use iso_c_binding, only : c_null_ptr
#endif /* LZ4 */

    implicit none

    integer        , intent(in)    :: compressor_id, compression_level
    type(c_ptr)    , intent(in)    :: input
    integer(kind=8), intent(in)    :: ilen
    type(c_ptr)    , intent(inout) :: buffer
    integer(kind=8), intent(inout) :: clen

#ifdef LZ4
    integer, dimension(14), target :: prefs = [5, 0, 1, 0, 0, 0, 0, 0, &
                                                           0, 0, 0, 0, 0, 0]
#endif /* LZ4 */
#ifdef LZMA
    integer(kind=8), target :: lsize
    integer                 :: ret
#endif /* LZMA */

!-------------------------------------------------------------------------------
!
    select case(compressor_id)
#ifdef ZSTD
    case(compressor_zstd)
      clen = zstd_compress(buffer, clen, input, ilen, compression_level)
      if (zstd_iserror(clen) /= 0) clen = 0
#endif /* ZSTD */
#ifdef LZ4
    case(compressor_lz4)
      prefs(5:6) = transfer(ilen, [ 0_4 ])
      prefs(9) = compression_level
      clen = lz4_compress(buffer, clen, input, ilen, c_loc(prefs))
      if (lz4_iserror(clen) /= 0) clen = 0
#endif /* LZ4 */
#ifdef LZMA
    case(compressor_lzma)
      lsize = 0
      ret = lzma_compress(compression_level, 4, c_null_ptr, &
                           input, ilen, buffer, c_loc(lsize), clen)
      if (ret > 0) then
        clen = 0
      else
        clen = lsize
      end if
#endif /* LZMA */
    case default
      clen = 0
    end select

!-------------------------------------------------------------------------------
!
  end subroutine compress
!
!===============================================================================
!
! subroutine CHECK_ENCODER:
! ------------------------
!
!   Subroutine sets the encoder ID based on the encoder name.
!
!   Arguments:
!
!     encoder_name   - the data encoder name;
!     encoder_id     - the data encoder id;
!
!===============================================================================
!
  subroutine check_encoder(encoder_name, encoder_id)

    implicit none

    character(len=*) , intent(in   ) :: encoder_name
    integer          , intent(  out) :: encoder_id

!-------------------------------------------------------------------------------
!
    select case(trim(adjustl(encoder_name)))
    case('shuffle', 'SHUFFLE')
      encoder_id = encoder_shuffle
    case('bytedelta', 'BYTEDELTA')
      encoder_id = encoder_bytedelta
    case default
      encoder_id = encoder_none
    end select

!-------------------------------------------------------------------------------
!
  end subroutine check_encoder
!
!===============================================================================
!
! function GET_ENCODER_NAME:
! -------------------------
!
!   Function returns the name of the data encoder.
!
!   Arguments:
!
!     encoder_id - the data encoder ID;
!
!===============================================================================
!
  character(len=16) function get_encoder_name(encoder_id)

    implicit none

    integer, intent(in) :: encoder_id

!-------------------------------------------------------------------------------
!
    select case(encoder_id)
    case(encoder_shuffle)
      get_encoder_name = 'shuffle'
    case(encoder_bytedelta)
      get_encoder_name = 'bytedelta'
    case default
      get_encoder_name = 'none'
    end select
    return

!-------------------------------------------------------------------------------
!
  end function get_encoder_name
!
!===============================================================================
!
! function IS_ENCODER_ON:
! ----------------------
!
!   The function determines if any data encoding is used.
!
!   Arguments:
!
!     encoder_id - the data encoder ID;
!
!===============================================================================
!
  logical function is_encoder_on(encoder_id)

    implicit none

    integer, intent(in) :: encoder_id

!-------------------------------------------------------------------------------
!
    is_encoder_on = encoder_id /= encoder_none
    return

!-------------------------------------------------------------------------------
!
  end function is_encoder_on
!
!===============================================================================
!
! subroutine ENCODE:
! -----------------
!
!   Subroutine encodes the input buffer for better data compression.
!
!   Arguments:
!
!     encoder_id - the data encoder ID;
!     item_size  - the size of each item in bytes;
!     bytes      - the length of the input data;
!     input_ptr  - the pointer to the input data;
!     output     - the output where the encoded data are written;
!
!===============================================================================
!
  subroutine encode(encoder_id, item_size, bytes, input_ptr, output)

    use iso_c_binding, only : c_ptr, c_f_pointer

    implicit none

    integer                          , intent(in  )  :: encoder_id
    integer(kind=4)                  , intent(in  )  :: item_size
    integer(kind=8)                  , intent(in  )  :: bytes
    type(c_ptr)                      , intent(in  )  :: input_ptr
    integer(kind=1), dimension(bytes), intent(inout) :: output

    integer(kind=1), dimension(:), pointer :: input

    integer(kind=8) :: i, j, m, n, item_count

!-------------------------------------------------------------------------------
!
    call c_f_pointer(input_ptr, input, [ bytes ])

    select case(encoder_id)
    case(encoder_shuffle)
      item_count = bytes / item_size

      i = 1
      j = item_count
      do m = 1, item_size
        output(i:j) = input(m:bytes:item_size)

        i = j + 1
        j = j + item_count
      end do

    case(encoder_bytedelta)
      item_count = bytes / item_size

      i = 1
      j = item_count
      do m = 1, item_size
        output(i:j) = input(m:bytes:item_size)

        do n = j, i + 1, -1
          output(n) = output(n) - output(n-1)
        end do

        i = j + 1
        j = j + item_count
      end do

    case default
      output(:) = input(:)
    end select

!-------------------------------------------------------------------------------
!
  end subroutine encode

!===============================================================================
!
end module compression
